# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pylammpsmpi.wrapper.base import LammpsConcurrent

__author__ = "Sarath Menon, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 28, 2020"


func_list = [
    "file",
    "extract_global",
    "extract_box",
    "extract_atom",
    "extract_fix",
    "extract_variable",
    "get_natoms",
    "set_variable",
    "reset_box",
    "generate_atoms",
    "set_fix_external_callback",
    "get_neighlist",
    "find_pair_neighlist",
    "find_fix_neighlist",
    "find_compute_neighlist",
    "get_neighlist_size",
    "get_neighlist_element_neighbors",
    "command",
    "gather_atoms",
    "scatter_atoms",
    "get_thermo",
    "extract_compute",
    "create_atoms",
]


prop_list = [
    "version",
    "natoms",
    "has_exceptions",
    "has_gzip_support",
    "has_png_support",
    "has_jpeg_support",
    "has_ffmpeg_support",
    "installed_packages",
]


command_list = [
    "angle_coeff",
    "angle_style",
    "atom_modify",
    "atom_style",
    "atom_style",
    "balance",
    "bond_coeff",
    "box",
    "bond_style",
    "boundary",
    "change_box",
    "clear",
    "comm_modify",
    "compute",
    "compute_modify",
    "create_atoms",
    "create_bonds",
    "create_box",
    "delete_atoms",
    "delete_bonds",
    "dielectric",
    "dihedral_coeff",
    "dihedral_style",
    "dimension",
    "displace_atoms",
    "dump",
    "dynamical_matrix",
    "fix",
    "fix_modify",
    "echo",
    "group2ndx",
    "ndx2group",
    "group",
    "hyper",
    "improper_coeff",
    "improper_style",
    "include",
    "info",
    "jump",
    "kim_init",
    "kim_interactions",
    "kim_query",
    "kim_param",
    "kspace_modify",
    "kspace_style",
    "label",
    "log",
    "message",
    "min_modify",
    "lattice",
    "mass",
    "minimize",
    "minimize/kk",
    "min_style",
    "molecule",
    "neb",
    "neb/spin",
    "next",
    "neighbor",
    "neigh_modify",
    "newton",
    "nthreads",
    "package",
    "pair_coeff",
    "pair_modify",
    "pair_style",
    "pair_write",
    "partition",
    "prd",
    "print",
    "python",
    "processors",
    "read",
    "read_data",
    "read_dump",
    "read_restart",
    "region",
    "replicate",
    "rerun",
    "reset_ids",
    "reset_timestep",
    "restart",
    "run",
    "run_style",
    "server",
    "set",
    "shell",
    "special_bonds",
    "suffix",
    "tad",
    "temper",
    "thermo",
    "thermo_modify",
    "thermo_style",
    "third_order",
    "timer",
    "timestep",
    "uncompute",
    "undump",
    "unfix",
    "units",
    "variable",
    "velocity",
    "write_coeff",
    "write_data",
    "write_dump",
    "write_restart",
]


thermo_list = [
    "step",
    "elapsed",
    "elaplong",
    "dt",
    "time",
    "cpu",
    "tpcpu",
    "spcpu",
    "cpuremain",
    "part",
    "timeremain",
    "atoms",
    "temp",
    "press",
    "pe",
    "ke",
    "etotal",
    "enthalpy",
    "evdwl",
    "ecoul",
    "epair",
    "ebond",
    "eangle",
    "edihed",
    "eimp",
    "emol",
    "elong",
    "etail",
    "vol",
    "density",
    "lx",
    "ly",
    "lz",
    "xlo",
    "xhi",
    "ylo",
    "yhi",
    "zlo",
    "zhi",
    "xy",
    "xz",
    "yz",
    "xlat",
    "ylat",
    "zlat",
    "bonds",
    "angles",
    "dihedrals",
    "impropers",
    "pxx",
    "pyy",
    "pzz",
    "pxy",
    "pxz",
    "pyz",
    "fmax",
    "fnorm",
    "nbuild",
    "ndanger",
    "cella",
    "cellb",
    "cellc",
    "cellalpha",
    "cellbeta",
    "cellgamma",
]


class LammpsLibrary:
    """
    Top level class which manages the lammps library provided by LammpsBase
    """

    def __init__(
        self,
        cores=1,
        oversubscribe=False,
        working_directory=".",
        client=None,
        mode="local",
        cmdargs=None,
    ):
        self.cores = cores
        self.working_directory = working_directory
        self.oversubscribe = oversubscribe
        self.client = client
        self.mode = mode
        self.lmp = LammpsConcurrent(
            cores=self.cores,
            oversubscribe=self.oversubscribe,
            working_directory=self.working_directory,
            cmdargs=cmdargs,
        )

    def __getattr__(self, name):
        """
        Try to run input as a lammps command
        """
        if name in func_list:

            def func_wrapper(*args, **kwargs):
                func = getattr(self.lmp, name)
                fut = func(*args, **kwargs)
                return fut.result()

            return func_wrapper

        elif name in thermo_list:
            fut = self.lmp.get_thermo(name)
            return fut.result()

        elif name in command_list:

            def command_wrapper(*args):
                args = [name] + list(args)
                cmd = " ".join([str(x) for x in args])
                fut = self.lmp.command(cmd)
                return fut.result()

            return command_wrapper

        elif name in prop_list:
            fut = getattr(self.lmp, name)
            return fut.result()

        else:
            raise AttributeError(name)

    def close(self):
        self.lmp.close()

    def __dir__(self):
        return (
            list(super().__dir__()) + func_list + thermo_list + command_list + prop_list
        )



from ._version import get_versions

from pylammpsmpi.wrapper.extended import LammpsLibrary
from pylammpsmpi.wrapper.concurrent import LammpsConcurrent
from pylammpsmpi.wrapper.base import LammpsBase


__all__ = [LammpsLibrary, LammpsConcurrent, LammpsBase]
__version__ = get_versions()["version"]


try:
    from pylammpsmpi.wrapper.ase import LammpsASELibrary

    __all__ += [LammpsASELibrary]
except ImportError:
    pass



import logging
import unittest

from ase.build import bulk
from ase.constraints import FixAtoms, FixedPlane, FixCom
import numpy as np

from pylammpsmpi import LammpsASELibrary, LammpsLibrary
from pylammpsmpi.wrapper.ase import (
    cell_is_skewed,
    get_species_symbols,
    get_structure_indices,
    get_lammps_indicies_from_ase_structure,
    set_selective_dynamics,
    UnfoldingPrism,
)


class TestLammpsASELibrary(unittest.TestCase):
    def test_static(self):
        lmp = LammpsASELibrary(
            working_directory=None,
            cores=1,
            comm=None,
            logger=logging.getLogger("TestStaticLogger"),
            log_file=None,
            library=LammpsLibrary(cores=2, mode="local"),
            diable_log_file=True,
        )
        structure = bulk("Al", cubic=True).repeat([2, 2, 2])
        lmp.interactive_lib_command(command="units lj")
        lmp.interactive_lib_command(command="atom_style atomic")
        lmp.interactive_lib_command(command="atom_modify map array")
        lmp.interactive_structure_setter(
            structure=structure,
            units="lj",
            dimension=3,
            boundary=" ".join(["p" if coord else "f" for coord in structure.pbc]),
            atom_style="atomic",
            el_eam_lst=["Al"],
            calc_md=False,
        )
        lmp.interactive_lib_command("pair_style lj/cut 6.0")
        lmp.interactive_lib_command("pair_coeff 1 1 1.0 1.0 4.04")
        lmp.interactive_lib_command(
            command="thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol"
        )
        lmp.interactive_lib_command(command="thermo_modify format float %20.15g")
        lmp.interactive_lib_command("run 0")
        self.assertTrue(
            np.all(np.isclose(lmp.interactive_cells_getter(), structure.cell.array))
        )
        self.assertTrue(
            np.isclose(lmp.interactive_energy_pot_getter(), -0.04342932384411341)
        )
        self.assertTrue(
            np.isclose(lmp.interactive_energy_tot_getter(), -0.04342932384411341)
        )
        self.assertTrue(np.isclose(np.sum(lmp.interactive_forces_getter()), 0.0))
        self.assertTrue(np.isclose(lmp.interactive_volume_getter(), 531.4409999999999))
        self.assertTrue(
            np.all(lmp.interactive_indices_getter() == [1] * len(structure))
        )
        self.assertEqual(lmp.interactive_steps_getter(), 0)
        self.assertEqual(lmp.interactive_temperatures_getter(), 0)
        self.assertTrue(
            np.isclose(
                np.sum(lmp.interactive_pressures_getter()), -0.015661731917941832
            )
        )
        self.assertEqual(np.sum(lmp.interactive_velocities_getter()), 0.0)
        self.assertTrue(np.isclose(np.sum(lmp.interactive_positions_getter()), 291.6))
        lmp.interactive_cells_setter(cell=1.01 * structure.cell.array)
        lmp.interactive_lib_command("run 0")
        self.assertTrue(
            np.all(
                np.isclose(lmp.interactive_cells_getter(), 1.01 * structure.cell.array)
            )
        )
        lmp.close()

    def test_static_with_statement(self):
        structure = bulk("Al").repeat([2, 2, 2])
        with LammpsASELibrary(
            working_directory=None,
            cores=2,
            comm=None,
            logger=None,
            log_file=None,
            library=None,
            diable_log_file=True,
        ) as lmp:
            lmp.interactive_lib_command(command="units lj")
            lmp.interactive_lib_command(command="atom_style atomic")
            lmp.interactive_lib_command(command="atom_modify map array")
            lmp.interactive_structure_setter(
                structure=structure,
                units="lj",
                dimension=3,
                boundary=" ".join(["p" if coord else "f" for coord in structure.pbc]),
                atom_style="atomic",
                el_eam_lst=["Al"],
                calc_md=False,
            )
            lmp.interactive_lib_command("pair_style lj/cut 6.0")
            lmp.interactive_lib_command("pair_coeff 1 1 1.0 1.0 4.04")
            lmp.interactive_lib_command(
                command="thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol"
            )
            lmp.interactive_lib_command(command="thermo_modify format float %20.15g")
            lmp.interactive_lib_command("run 0")
            self.assertTrue(
                np.isclose(lmp.interactive_energy_pot_getter(), -0.3083820387630098)
            )
            self.assertTrue(
                np.isclose(lmp.interactive_energy_tot_getter(), -0.3083820387630098)
            )
            self.assertTrue(np.isclose(np.sum(lmp.interactive_forces_getter()), 0.0))
            self.assertTrue(
                np.isclose(lmp.interactive_volume_getter(), 132.86024999999998)
            )
            self.assertTrue(
                np.all(lmp.interactive_indices_getter() == [1] * len(structure))
            )
            self.assertEqual(lmp.interactive_steps_getter(), 0)
            self.assertEqual(lmp.interactive_temperatures_getter(), 0)
            self.assertTrue(
                np.isclose(
                    np.sum(lmp.interactive_pressures_getter()), -0.00937227406237915
                )
            )
            self.assertEqual(np.sum(lmp.interactive_velocities_getter()), 0.0)


class TestASEHelperFunctions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.structure_skewed = bulk("Al").repeat([2, 2, 2])
        cls.structure_cubic = bulk("Al", cubic=True).repeat([2, 2, 2])

    def test_get_species_symbols(self):
        self.assertEqual(get_species_symbols(structure=self.structure_cubic), ["Al"])

    def test_get_structure_indices(self):
        indicies = get_structure_indices(structure=self.structure_cubic)
        self.assertEqual(len(indicies), len(self.structure_cubic))
        self.assertEqual(len(set(indicies)), 1)
        self.assertEqual(set(indicies), {0})

    def test_cell_is_skewed(self):
        self.assertTrue(cell_is_skewed(cell=self.structure_skewed.cell))
        self.assertFalse(cell_is_skewed(cell=self.structure_cubic.cell))

    def test_get_lammps_indicies_from_ase_structure(self):
        indicies = get_lammps_indicies_from_ase_structure(
            structure=self.structure_cubic, el_eam_lst=["Al", "H"]
        )
        self.assertEqual(len(indicies), len(self.structure_cubic))
        self.assertEqual(len(set(indicies)), 1)
        self.assertEqual(set(indicies), {1})

    def test_unfolding_prism_cubic(self):
        prism = UnfoldingPrism(self.structure_cubic.cell.array)
        self.assertEqual(
            prism.get_lammps_prism_str(),
            ("8.1000000000", "8.1000000000", "8.1000000000", "0E-10", "0E-10", "0E-10"),
        )
        self.assertTrue(
            np.all(
                np.isclose(
                    prism.pos_to_lammps(position=[[1.0, 1.0, 1.0]]),
                    np.array([1.0, 1.0, 1.0]),
                )
            )
        )

    def test_unfolding_prism_skewed(self):
        prism = UnfoldingPrism(self.structure_skewed.cell.array)
        self.assertTrue(
            np.all(
                np.isclose(
                    [np.abs(float(s)) for s in prism.get_lammps_prism_str()],
                    [
                        5.7275649276,
                        4.9602167291,
                        4.6765371804,
                        2.8637824638,
                        2.8637824638,
                        1.6534055764,
                    ],
                )
            )
        )
        self.assertTrue(
            np.all(
                np.isclose(
                    prism.pos_to_lammps(position=[[1.0, 1.0, 1.0]]),
                    np.array([1.41421356, 0.81649658, 0.57735027]),
                )
            )
        )


class TestConstraints(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        structure = bulk("Cu", cubic=True)
        structure.symbols[2:] = "Al"
        cls.structure = structure

    def test_selective_dynamics_mixed_calcmd(self):
        atoms = self.structure.copy()
        c1 = FixAtoms(indices=[atom.index for atom in atoms if atom.symbol == "Cu"])
        c2 = FixedPlane(
            [atom.index for atom in atoms if atom.symbol == "Al"],
            [1, 0, 0],
        )
        atoms.set_constraint([c1, c2])
        control_dict = set_selective_dynamics(structure=atoms, calc_md=True)
        self.assertEqual(len(control_dict), 6)
        self.assertTrue(control_dict["group constraintxyz"], "id 1 2")
        self.assertTrue(
            control_dict["fix constraintxyz"], "constraintxyz setforce 0.0 0.0 0.0"
        )
        self.assertTrue(control_dict["velocity constraintxyz"], "set 0.0 0.0 0.0")
        self.assertTrue(control_dict["group constraintx"], "id 3 4")
        self.assertTrue(
            control_dict["fix constraintx"], "constraintx setforce 0.0 NULL NULL"
        )
        self.assertTrue(control_dict["velocity constraintx"], "set 0.0 NULL NULL")

    def test_selective_dynamics_mixed(self):
        atoms = self.structure.copy()
        c1 = FixAtoms(indices=[atom.index for atom in atoms if atom.symbol == "Cu"])
        c2 = FixedPlane(
            [atom.index for atom in atoms if atom.symbol == "Al"],
            [1, 0, 0],
        )
        atoms.set_constraint([c1, c2])
        control_dict = set_selective_dynamics(structure=atoms, calc_md=False)
        self.assertEqual(len(control_dict), 4)
        self.assertTrue(control_dict["group constraintxyz"], "id 1 2")
        self.assertTrue(
            control_dict["fix constraintxyz"], "constraintxyz setforce 0.0 0.0 0.0"
        )
        self.assertTrue(control_dict["group constraintx"], "id 3 4")
        self.assertTrue(
            control_dict["fix constraintx"], "constraintx setforce 0.0 NULL NULL"
        )

    def test_selective_dynamics_single_fix(self):
        atoms = self.structure.copy()
        c1 = FixAtoms(indices=[atom.index for atom in atoms if atom.symbol == "Cu"])
        atoms.set_constraint(c1)
        control_dict = set_selective_dynamics(structure=atoms, calc_md=False)
        self.assertEqual(len(control_dict), 2)
        self.assertTrue(control_dict["group constraintxyz"], "id 1 2")
        self.assertTrue(
            control_dict["fix constraintxyz"], "constraintxyz setforce 0.0 0.0 0.0"
        )

    def test_selective_dynamics_errors(self):
        atoms = self.structure.copy()
        atoms.set_constraint(FixCom())
        with self.assertRaises(ValueError):
            set_selective_dynamics(structure=atoms, calc_md=False)

    def test_selective_dynamics_wrong_plane(self):
        atoms = self.structure.copy()
        atoms.set_constraint(
            FixedPlane(
                [atom.index for atom in atoms if atom.symbol == "Al"],
                [2, 1, 0],
            )
        )
        with self.assertRaises(ValueError):
            set_selective_dynamics(structure=atoms, calc_md=False)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import unittest
import numpy as np
import os
from pylammpsmpi import LammpsBase


class TestLammpsBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.execution_path = os.path.dirname(os.path.abspath(__file__))
        cls.citation_file = os.path.join(cls.execution_path, "citations.txt")
        cls.lammps_file = os.path.join(cls.execution_path, "in.simple")
        cls.lmp = LammpsBase(
            cores=1,
            oversubscribe=False,
            working_directory=".",
            cmdargs=["-cite", cls.citation_file],
        )
        cls.lmp.file(cls.lammps_file)

    @classmethod
    def tearDownClass(cls):
        cls.lmp.close()

    def test_file_not_found(self):
        lmp = LammpsBase(
            cores=1,
            oversubscribe=False,
            working_directory=".",
            cmdargs=["-cite", self.citation_file],
        )
        with self.assertRaises(FileNotFoundError):
            lmp.file("file_does_not_exist.txt")

    def test_extract_atom(self):
        f = self.lmp.extract_atom("f")
        self.assertEqual(len(f), 256)
        self.assertEqual(np.round(f[0][0], 2), -0.26)

        ids = self.lmp.extract_atom("id")
        self.assertEqual(len(ids), 256)

    def test_extract_compute_global(self):
        compute_temp = self.lmp.extract_compute("1", 0, 0)
        self.assertIsInstance(compute_temp, float)

    def test_extract_compute_per_atom(self):
        compute_ke_atom = self.lmp.extract_compute("ke", 1, 1)
        self.assertEqual(len(compute_ke_atom), 256)

    def test_gather_atoms(self):
        f = self.lmp.gather_atoms("f")
        self.assertEqual(len(f), 256)
        # this checks if info was gathered from
        # all processors
        self.assertEqual(np.round(f[-22][0], 2), 0.31)

        ids = self.lmp.extract_atom("id")
        self.assertEqual(len(ids), 256)
        self.assertEqual(self.lmp.get_natoms(), 256)
        self.assertEqual(self.lmp.natoms, 256)

    def test_extract_fix(self):
        x = self.lmp.extract_fix("2", 0, 1, 1)
        self.assertEqual(np.round(x, 2), -2.61)

    def test_extract_variable(self):
        x = self.lmp.extract_variable("tt", "all", 0)
        self.assertEqual(np.round(x, 2), 1.13)
        x = self.lmp.extract_variable("fx", "all", 1)
        self.assertEqual(len(x), 256)
        self.assertEqual(np.round(x[0], 2), -0.26)

    def test_scatter_atoms(self):
        f = self.lmp.gather_atoms("f")
        val = np.random.randint(0, 100)
        f[1][0] = val
        self.lmp.scatter_atoms("f", f)
        f1 = self.lmp.gather_atoms("f")
        self.assertEqual(f1[1][0], val)

        f = self.lmp.gather_atoms("f", ids=[1, 2])
        val = np.random.randint(0, 100)
        f[1][1] = val
        self.lmp.scatter_atoms("f", f, ids=[1, 2])
        f1 = self.lmp.gather_atoms("f", ids=[1, 2])
        self.assertEqual(f1[1][1], val)

    def test_extract_box(self):
        box = self.lmp.extract_box()
        self.assertEqual(len(box), 7)

        self.assertEqual(box[0][0], 0.0)
        self.assertEqual(np.round(box[1][0], 2), 6.72)

        self.lmp.command("delete_atoms group all")
        self.lmp.reset_box([0.0, 0.0, 0.0], [8.0, 8.0, 8.0], 0.0, 0.0, 0.0)
        box = self.lmp.extract_box()
        self.assertEqual(box[0][0], 0.0)
        self.assertEqual(np.round(box[1][0], 2), 8.0)
        self.lmp.command("clear")
        self.lmp.file(self.lammps_file)

    def test_cmdarg_options(self):
        self.assertTrue(os.path.isfile(self.citation_file))

    def test_version(self):
        self.assertTrue(self.lmp.version in [20220623, 20230802, 20231121, 20240207])

    def test_extract_global(self):
        self.assertEqual(
            self.lmp.extract_global(name="boxhi"),
            [6.718384765530029, 6.718384765530029, 6.718384765530029],
        )
        self.assertEqual(self.lmp.extract_global(name="boxlo"), [0.0, 0.0, 0.0])

    def test_properties(self):
        self.assertEqual(self.lmp.has_exceptions, True)
        self.assertEqual(self.lmp.has_gzip_support, True)
        self.assertEqual(self.lmp.has_png_support, True)
        self.assertEqual(self.lmp.has_jpeg_support, True)
        self.assertEqual(self.lmp.has_ffmpeg_support, False)

    def test_get_thermo(self):
        self.assertEqual(float(self.lmp.get_thermo("temp")), 1.1298532212880312)


if __name__ == "__main__":
    unittest.main()
